#
# pyl3dgecomm v0.2
#
# Copyright (c) 2012 Centre for Advanced Internet Architectures,
# Swinburne University of Technology.
#
# Authors:    Grenville Armitage  (garmitage@swin.edu.au)
#             Chris Holman        (6963420@student.swin.edu.au)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.


# TODO:
#	- Action token generation should be random

import socket

# *** split_addr_and_port()
#
# Function to parse <server>:<port> and return w.x.y.z and port
# (converting <server> from domain name to numeric as needed)
#
def splitAddressPort(server,def_port):
 "Return w.x.y.z (string) and port (int) when passed a string <server>:<port>"
 "<server> may be a domain name or w.x.y.z literal"
 "2nd parameter is default port to return if there's no ':port' in first parameter"
 hostname_port = []
 # Remove any leading spaces from the hostname, then
 # split again at the ":" separating name and port, before
 # issuing gethostbyname
 hostname_port = (server.split()[0]).split(":")
 hostname = hostname_port[0]
 if len(hostname_port) > 1:
  hostport = int(hostname_port[1])
 else:
  hostport = int(def_port)
 ipaddr = socket.gethostbyname(hostname)
 return [ipaddr, hostport]


# *** create_socket()
#
# Function to create a local UDP socket 'attached' to a specified
# localaddr:port, with which we can sendto() and recvfrom() later.
#
# We return udp socket.
#
def createSocket(localaddr,port):
  "Create UDP socket bound to server:port"
  udpsock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM)
  try:
    udpsock.bind((localaddr,port))
  except socket.error as err:
    if (debugflag == 1):
      print ("createSocket(): Couldn't bind socket, bailing out and pulling the ripcord!")
    raise SystemExit

  return udpsock
 
 
# *** request_token()
#
# Function to contact a l3dgeworld server and request a token
# to use in future communication. We assume a pre-existing socket
# is passed as sole parameter.
#
# We return the token.
#
def requestToken(server,port,udpsock):
  "Contact target server:port and returns a l3dgeworld token"
  
  # Initialise token, but this will eventually be overwritten
  token = b"<blanktoken>"
  
  # Construct control message with 0xFFFFFF preamble.
  gettoken_msg = b"\xFF\xFF\xFF\xFFgettoken\0"
  
  if debugflag == 1:
    print ("requestToken(): Sending %s to the server" % (gettoken_msg[4:]) )
  
  # Send the actual message as connectionless message
  udpsock.sendto(gettoken_msg,(server,port))

  # Set 5 second timeout for socket read
  udpsock.settimeout(5)
  try:
    udpreply = udpsock.recv(1024)
  except socket.error as err:
    if debugflag == 1:
      print ("requestToken(): socket.recv() timed out")
    return token

  if debugflag == 1:
    print ("requestToken(): Received reply of %s" % udpreply)
  token = udpreply[15:31] # 15 is \xFF\xFF\xFF\xFFrequesttoken
  return token
 
 
 

class l3dgecomm:
  
  # server_str: ip/host and optional ":port"
  # default_serv_port: port to use if not specified in server_Str
  # local_str: local IP to use
  def __init__(self, server_str,default_serv_port,local_str="",accept_actions=False):
    "Establish a socket to target l3dgeworld server and intialise the class"
    
    self.debugflag = 0
    
    # action_queue is a list of actions waiting 
    # Actions are added on the queue, and removed when an ack comes
    # Every n milliseconds, the action is sent
    self.accept_actions = accept_actions
    self.action_ack = 0
    self.action_token = None
    
    [self.l3_serv,self.l3_serv_port] = splitAddressPort(server_str,default_serv_port)
    if self.debugflag == 1:
      print ("Target l3dgeworld server is: %s" % ([self.l3_serv,self.l3_serv_port]))

    # Initialise socket
    try:
        self.l3_sock = createSocket(local_str,27961) # XXX: set this to 0 again
    except socket.error as err:
      if self.debugflag == 1:
        print ("initialiseComms(): Couldn't create connection: %s" % (err))
      raise SystemExit

    # Initialise the connection to the l3dge server, get a token
    self.l3_token = requestToken(self.l3_serv,self.l3_serv_port,self.l3_sock)
    if self.debugflag == 1:
      print ("initialiseComms(): Received token is %s " % (self.l3_token))
    

  
  

  # *** checkIncoming()
  #
  # Check to see if there's any "invalidtoken" message waiting for us in
  # the inbound queue from a previous failed tranmission attempt. If there is,
  # and it matches the current token, we request a new token and modify
  # the l3state list. We return the (possibly modified) list to the caller.
  #
  def checkIncoming(self):
    """ Process the queue of incoming packets """

    # Set 0.1ms timeout for socket read
    # (Note this has the side-effect of slightly rate-limiting our ability to
    # hammer the l3dgeworld server with command messages if we're waiting a little
    # each time to check for tokens)
    self.l3_sock.settimeout(0.0001)
    # Disable waiting for a return packet
    #self.l3_sock.setblocking(0)
    gotInvalid = 0
    

    while True:
      udpreplytext = b""
      # Check if a packet is in the queue. If it is, get it. 
      # IF it's not, it'll throw an exception, and we break
      try:
        [udpreply,addr] = self.l3_sock.recvfrom(1024)
        
      except socket.timeout as err:
        # Getting to here means there's no message waiting
        if self.debugflag == 1:
          print ("checkIncoming(): socket.recv() timed out -- that's good!")
        self.l3_sock.settimeout(5)
        break
      
      # Check packet source address
      if addr[0] != self.l3_serv:
        if self.debug:
            print  ("Received message from unauthorised IP address: " + addr[0] )
        continue
      
      # If we've voided our current token, flush all of the current packet queue
      if self.l3_token == None:
        continue

      if self.debugflag == 1:
        print ("checkIncoming(): There was a message waiting for us,,,, %s" % udpreply[4:])
      
      udpreplytext = udpreply[4:]
      
      # Check for invalidtoken messages. Erasing the current token serves as a signal to
      # other functions that they cannot send messages for a bit (but we're not multithreaded,
      # so it's not really an issue)
      if udpreplytext.find(b"invalidtoken") == 0:
        self.l3_token = None
        gotInvalid = 1
        continue
      
      if udpreplytext.find(b"action") == 0 and self.accept_actions == True:
        # Check tokens
        bits = udpreplytext.split("~")
        if bits[1] != self.action_token:
            if self.debugflag == 1:
                print("Invalid action token received from server")
            continue
        ackNum = bits[2]
        entId = bits[3]
        toolId = bits[4]
        entDesc = bits[5]
        
        #ack the action
        replyMsg = b"\xFF\xFF\xFF\xFFack " + self.l3_token + " " + ackNum
        self.l3_sock.sendto(replyMsg, addr)
        
        # Append action to action queue
        #self.action_queue.append( [entId, toolId, entDesc] )
        #print(entId + " " + toolId + " " + entDesc)
        return [entId, toolId, entDesc] 
        
      if udpreplytext.find(b"gettoken") == 0 and self.accept_actions == True:
        # Generate new token
        self.action_token = b"1234567812345678"
        
        # Send token to server
        replyMsg = b"\xFF\xFF\xFF\xFFoutputtoken " + self.l3_token + b" " + self.action_token
        self.l3_sock.sendto(replyMsg, addr)
        continue
        
      
    
    
    # We need to wait for the l3dgetoken to come back
    #self.l3_socket.setblocking(1)
    #self.l3_socket.settimeout(5)
    if gotInvalid:
      while True:
        try:
          udpreply = self.l3_sock.recv(1024)
        except socket.timeout as err:
          # Getting to here means there's no more messages waiting
          if self.debugflag == 1:
            print ("checkIncoming(): socket.recv() timed out -- now let us request a new token!")
          # so we go ahead and request a new token
          self.l3_token = requestToken(self.l3_serv,self.l3_serv_port,self.l3_sock)
          return

    return False# Never get to here anyway

    
    
  
  # *** sendEntityUpdate()
  #
  # Function to send a message
  #
  # \param entityid is the numeric id of the entity we're updating
  # \param metric is the metric to update (integer)
  # \param field is the field of the metric to update (character "n", "v" or "r")
  # \param value is a string containing the data we wish to set.
  #
  def sendEntityUpdate(self,entityid,metric,field,value):
    "Send a request to update an entity on target l3dgeworld server"

    if self.l3_token == None:
        return False
    
    
    # Now construct message to be sent to l3dgeworld server

    
    # Message is in the form:
    #    "l3dge <token> ~<entity_id>~<characteristicnumber>~<field>~<value>~"
    # where entity_id is a three digit field.

    # Create threedigit entity number (with leading zeros if required)
    entityid_str = '%03i' % int(entityid)

    # Construct control message with 0xFFFFFF preamble.
    updatemsg = b"\xFF\xFF\xFF\xFFl3dge "+self.l3_token+b" ~"+entityid_str.encode('latin-1')+b"~"+str(metric).encode('latin-1')+b"~"+field.encode('latin-1')+b"~"+value.encode('latin-1')+b"~\0"

    if self.debugflag == 1:
      print ("sendEntityUpdate(): Sending %s" % (updatemsg[4:]))

    # Send the actual message as connectionless message
    self.l3_sock.sendto(updatemsg,(self.l3_serv,self.l3_serv_port))
    return True

  # *** triggerEntityCoordUpdate()
  #
  # Send a message to a specified l3dgeworld server to update the coordinates
  # of a specific entity
  #
  def triggerEntityCoordUpdate(self,entityid):
    "Send a request to update an entity's coordinates on target l3dgeworld server"

    if self.l3_token == None:
        return False
    
    # Now construct message to be sent to l3dgeworld server
    # Message is in the form:
    #    " move <token> ~<entity_id>~"
    # where entity_id is a three digit field.
      
    # Create threedigit entity number (with leading zeros if required)
    entityid_str = '%03i' % int(entityid)
    
    # Construct control message with 0xFFFFFF preamble.
    # (NOTE: There's a BUG in the l3dgeworld 2.3 documentation -- the "move" text
    # MUST be preceded by a space in addition to be trailed by a space.)
    updatemsg = b"\xFF\xFF\xFF\xFF move "+self.l3_token+b" ~"+entityid_str.encode('latin-1')+b"~\0"

    if self.debugflag == 1:
      print ("triggerEntityCoordUpdate(): Sending %s" % (updatemsg[4:]))

    # Send the actual message as connectionless message
    self.l3_sock.sendto(updatemsg,(self.l3_serv,self.l3_serv_port))
    return True
    
  # *** triggerMultiEntityCoordUpdate()
  #
  # Send a message to a specified l3dgeworld server to update the coordinates
  # of two or more entities.
  #
  def triggerMultiEntityCoordUpdate(self,entityids):
    "Send a request to update two or more entity coordinates on target l3dgeworld server"

    if self.l3_token == None:
        return False
    
    # Now construct message to be sent to l3dgeworld server
    # Message is in the form:
    #    " move <token> ~<entity_id>~"
    # where entity_id is a three digit field.

    entityid_str =""

    for x in entityids:
      # Create threedigit entity number (with leading zeros if required)
      entityid_str = "~".join((entityid_str, '%03i' % int(x)))

    # Construct control message with 0xFFFFFF preamble.
    # (NOTE: There's a BUG in the l3dgeworld 2.3 documentation -- the "move" text
    # MUST be preceded by a space in addition to be trailed by a space.)
    updatemsg = b"\xFF\xFF\xFF\xFF move "+self.l3_token+b" "+entityid_str.encode('latin-1')+b"~\0"

    if self.debugflag == 1:
      print ("triggerMultiEntityCoordUpdate(): Sending %s" % (updatemsg[4:]))

    # Send the actual message as connectionless message
    self.l3_sock.sendto(updatemsg,(self.l3_serv,self.l3_serv_port))

    return True
   



  # *** sendFeedback()
  #
  # Send a text message to a specified l3dgeworld server which will be re-transmitted
  # as-is to all connected clients
  #
  def sendFeedback(self,message):
    "Send a message to target l3dgeworld server for all clients"

    if self.l3_token == None:
        return False    
    
    # Now construct message to be sent to l3dgeworld server
    # Message is in the form:
    #    "feedback <message>"
    # where entity_id is a three digit field.

    # Construct control message with 0xFFFFFF preamble.
    updatemsg = "\xFF\xFF\xFF\xFFfeedback "+message+"\0"

    if self.debugflag == 1:
      print ("sendFeedback(): Sending %s" % (updatemsg[4:]))

    # Send the actual message as connectionless message
    self.l3_sock.sendto(updatemsg,(self.l3_serv,self.l3_serv_port))

    return True #l3state


  # *** closeSocket()
  #
  # Close a previously-opened socket, just to tidy up
  #
  def closeSocket(self):
    "Close a previously-opened socket, just to tidy up"

    #s = l3state[2]
    self.l3_sock.close()
    self.delete
    return

# ***************** "Constants" that may be useful to other programs
#
# Use them as "pyl3dgecomm.X" where X is the constant's name
#

version = "0.1"

# Set this to higher values for more verbose output from each method/function
debugflag = 0

METRIC_ONE=    (1)
METRIC_TWO=    (2)
METRIC_THREE =  (3)
METRIC_FOUR =   (4)
METRIC_FIVE =   (5)
METRIC_SIX =    (6)
METRIC_SEVEN =  (7)
METRIC_EIGHT =  (8)
METRIC_NINE =   (9)
METRIC_TEN =    (10)
METRIC_ZERO =   (0)

METRIC_SPIN =  METRIC_ONE
METRIC_BOUNCE_HEIGHT =  METRIC_TWO
METRIC_SCALE =   METRIC_THREE
METRIC_ROLL = METRIC_FOUR
METRIC_COLOR = METRIC_FIVE
METRIC_BOUNCE_RATE  =  METRIC_SIX 
METRIC_SOUND =    METRIC_SEVEN
METRIC_ALT_SKIN = METRIC_EIGHT
METRIC_ALT_MODEL = METRIC_NINE 
METRIC_UNUSED =   METRIC_TEN
METRIC_INFO =  METRIC_ZERO


FIELD_NAME =  "n"
FIELD_VALUE = "v"
FIELD_RATE =  "r"



